# $Id: snmp.tcl,v 1.17 2001/04/23 13:47:27 gdmr Exp $
#
# Copyright 2000 Division of Informatics, University of Edinburgh
# 
# Permission to use, copy, modify, distribute, and sell this software and its
# documentation for any purpose is hereby granted without fee, provided that
# the above copyright notice appear in all copies and that both that
# copyright notice and this permission notice appear in supporting
# documentation, and that the name of the University of Edinburgh not be used
# in advertising or publicity pertaining to distribution of the software
# without specific, written prior permission.  The University of Edinburgh
# makes no representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied warranty.
#
# SNMP-based procedures to query the switches.  Cache the results
# where appropriate.
#

set mibList {3cVlan.mib 3cGeneric.mib 3cStackConfig.mib}

####################
# Initialisation
####################

package require Tnm 3.0
namespace import Tnm::snmp
namespace import Tnm::mib

puts "Loading mibs: $mibList"
foreach m $mibList {
   mib load [file join [file dirname $argv0] $m]
}

proc snmpInit {host community} {
   global snmpS
   if [info exists snmpS] {
      puts "Can't re-snmpInit"
   } else {
      set snmpS [snmp generator -address $host -community $community]
   }
}

proc snmpInit2 {host community} {
   global snmp2
   if [info exists snmp2] {
      puts "Can't re-snmpInit2"
   } else {
      set snmp2 [snmp generator -address $host -community $community]
   }
}

####################
# System group
####################

proc snmpDescr {} {
   global snmpS
   global snmpSysDescr
   if [info exists snmpSysDescr] {
      set snmpSysDescr
   } else {
      set vb "SNMPv2-MIB!sysDescr.0"
      set vb [$snmpS get $vb]
      eval set snmpSysDescr [snmp value $vb]
   }
}

proc snmpUpTime {} {
   # Don't cache this one
   global snmpS
   set vb "SNMPv2-MIB!sysUpTime.0"
   set vb [$snmpS get $vb]
   expr {[snmp value $vb] / 100}
}

proc snmpUpTime2 {} {
   # Don't cache this one either
   global snmp2
   set vb "SNMPv2-MIB!sysUpTime.0"
   set vb [$snmp2 get $vb]
   expr {[snmp value $vb] / 100}
}

proc snmpLocation {} {
   global snmpS
   global snmpSysLocation
   if [info exists snmpSysLocation] {
      set snmpSysLocation
   } else {
      set vb "SNMPv2-MIB!sysLocation.0"
      set vb [$snmpS get $vb]
      eval set snmpSysLocation [snmp value $vb]
   }
}

####################
# Interfaces
####################

proc snmpEtherIfIndices {} {
   global snmpS
   set ifRoot [mib oid "IF-MIB!ifIndex"]
   set vbl "$ifRoot.0
            IF-MIB!ifType.0"
   while {1} {
      set vbl [$snmpS getnext $vbl]
      if {! [mib subtree $ifRoot [snmp oid $vbl 0]]} break
      if {[snmp value $vbl 1] == "ethernetCsmacd"} {
         lappend indices [snmp value $vbl 0]
      }
   }
   # Don't need to sort, they'll be in numeric order already
   set indices
}

proc snmpIfType {i} {
   global snmpIfTypes
   global snmpS
   if {[info exists snmpIfTypes($i)]} {
      set snmpIfTypes($i)
   } else {
      set vb "IF-MIB!ifType.$i"
      set vb [$snmpS get $vb]
      set snmpIfTypes($i) [snmp value $vb]
   }
}

proc snmpIfDescr {i} {
   global snmpIfDescrs
   global snmpS
   if {[info exists snmpIfDescrs($i)]} {
      set snmpIfDescrs($i)
   } else {
      set vb "IF-MIB!ifDescr.$i"
      set vb [$snmpS get $vb]
      eval set snmpIfDescrs($i) [snmp value $vb]
   }
}

proc snmpIfUp {i} {
   global snmpS
   set vb "IF-MIB!ifOperStatus.$i"
   set vb [$snmpS get $vb]
   if {[snmp value $vb] == "up"} {
      return 1
   } else {
      return 0
   }
}

proc snmpIfSpeed {i} {
   global snmpS
   set vb "IF-MIB!ifSpeed.$i"
   set vb [$snmpS get $vb]
   snmp value $vb
}

proc snmpIfLastChange {i} {
   global snmpS
   set vb "IF-MIB!ifLastChange.$i"
   set vb [$snmpS get $vb]
   snmp value $vb
}

proc snmpIfStats {i} {
   global snmpS
   set vbl "IF-MIB!ifInOctets.$i
            IF-MIB!ifInUcastPkts.$i
            IF-MIB!ifInErrors.$i
            IF-MIB!ifOutOctets.$i
            IF-MIB!ifOutUcastPkts.$i
            IF-MIB!ifOutErrors.$i"
   set vbl [$snmpS get $vbl]
   snmp value $vbl
}

####################
# Interface stacking
####################

proc snmpGrabIfStacking {} {
   global snmpStackUp
   global snmpStackDown
   global snmpS
   set stackRoot [mib oid "IF-MIB!ifStackStatus"]
   set stackMatch "^$stackRoot\\.(.*)\\.(.*)$"
   set vb $stackRoot.0
   while {1} {
      set vb [$snmpS getnext $vb]
      if {! [mib subtree $stackRoot [snmp oid $vb]]} break
      if {[snmp value $vb] == "active"} {
         if [regexp $stackMatch [snmp oid $vb] all upper lower] {
            lappend snmpStackUp($lower) $upper
            lappend snmpStackDown($upper) $lower
         } else {
            puts stderr "snmpGrabIfStacking: no regexp match for $vb"
         }
      }
   }
}

proc snmpAreStackedAbove {} {
   global snmpStackUp
   lsort [array names snmpStackUp]
}

proc snmpAreStackedBelow {} {
   global snmpStackDown
   lsort [array names snmpStackDown]
}

proc snmpStackAbove {i} {
   global snmpStackUp
   if [info exists snmpStackUp($i)] {
      set snmpStackUp($i)
   } else {
      return ""
   }
}

proc snmpStackBelow {i} {
   global snmpStackDown
   if [info exists snmpStackDown($i)] {
      set snmpStackDown($i)
   } else {
      return ""
   }
}

####################
# VLANs
####################

proc snmpVlanIndices {} {
   global snmpS
   set vlanRoot [mib oid "a3ComVlanGlobalMappingIfIndex"]
   set vb $vlanRoot.0
   while {1} {
      set vb [$snmpS getnext $vb]
      if {! [mib subtree $vlanRoot [snmp oid $vb]]} break
      lappend indices [snmp value $vb]
   }
   set indices
}

proc snmpVlanDescrTag {i} {
   global vlanDescrTags
   global snmpS
   if {! [info exists vlanDescrTags($i)]} {
      set vbl "a3ComVlanIfDescr.$i
               a3ComVlanIfGlobalIdentifier.$i"
      set vbl [$snmpS get $vbl]
      set vlanDescrTags($i) [snmp value $vbl]
   }
   set vlanDescrTags($i)
}

proc snmpVlanDescr {i} {
   lindex [snmpVlanDescrTag $i] 0
}

proc snmpVlanTag {i} {
   lindex [snmpVlanDescrTag $i] 1
}

####################
# Encapsulation
####################

proc snmpEncapsulationIndices {} {
   global encapsIndices
   global snmpS
   set encapsRoot [mib oid "a3ComVlanEncapsIfStatus"]
   set encapsMatch "^$encapsRoot\\.(.*)$"
   set vb $encapsRoot.0
   while {1} {
      set vb [$snmpS getnext $vb]
      set oid [snmp oid $vb]
      if {! [mib subtree $encapsRoot $oid]} break
      if {[snmp value $vb] == "active"} {
         if [regexp $encapsMatch $oid all index] {
            lappend encapsIndices $index
         } else {
            puts stderr "snmpEncapsulationIndices: no regexp match for $vb"
         }
      }
   }
   # Should be sorted already
   set encapsIndices
}

proc snmpEncapsTagType {i} {
   global encapsTagTypes
   global snmpS
   if {! [info exists encapsTagTypes($i)]} {
      set vbl "a3ComVlanEncapsIfTag.$i
               a3ComVlanEncapsIfType.$i"
      set vbl [$snmpS get $vbl]
      set encapsTagTypes($i) [snmp value $vbl]
   }
   set encapsTagTypes($i)
}

proc snmpEncapsTag {i} {
   lindex [snmpEncapsTagType $i] 0
}

proc snmpEncapsType {i} {
   lindex [snmpEncapsTagType $i] 1
}

####################
# MAU
####################

proc snmpMauAutoType {i} {
   global snmpMauAutoTypes
   global snmpS
   if {! [info exists snmpMauTypes($i)]} {
      set vbl "MAU-MIB!ifMauType.$i
               MAU-MIB!ifMauAutoNegAdminStatus.$i"
      set vbl [$snmpS getnext $vbl]
      set snmpMauAutoTypes($i) [snmp value $vbl]
   }
   set snmpMauAutoTypes($i)
}

proc snmpMauType {i} {
   lindex [snmpMauAutoType $i] 0
}

proc snmpMauAuto {i} {
   lindex [snmpMauAutoType $i] 1
}

proc snmpMauAvailableExits {i} {
   global snmpS
   set vb "MAU-MIB!ifMauMediaAvailableStateExits.$i"
   set vb [$snmpS getnext $vb]
   snmp value $vb
}

####################
# Stack Config
####################

proc snmpStackConfigList {u} {
   global snmpS
   set vbl "A3COM0017-STACK-CONFIG!stackUnitDesc.$u
            A3COM0017-STACK-CONFIG!stackUnitName.$u
            A3COM0017-STACK-CONFIG!stackUnitPromVersion.$u
            A3COM0017-STACK-CONFIG!stackUnitHWVersion.$u
            A3COM0017-STACK-CONFIG!stackUnitSWVersion.$u
            A3COM0017-STACK-CONFIG!stackUnitSerialNumber.$u
            A3COM0017-STACK-CONFIG!stackUnitAddress.$u"
   set vbl [$snmpS get $vbl]
   snmp value $vbl
}

##########
# Bridge/STP
##########

proc snmpBridgeBasePorts {} {
   global snmpS
   global basePortToIfIndex
   global ifIndexToBasePort
   set bpRoot [mib oid "BRIDGE-MIB!dot1dBasePort"]
   set vbl "$bpRoot.0
            BRIDGE-MIB!dot1dBasePortIfIndex.0"
   while {1} {
      set vbl [$snmpS getnext $vbl]
      if {! [mib subtree $bpRoot [snmp oid $vbl 0]]} break
      set basePortToIfIndex([snmp value $vbl 0]) [snmp value $vbl 1]
      set ifIndexToBasePort([snmp value $vbl 1]) [snmp value $vbl 0]
   }
}

proc snmpBridgeBasePortIf {bp} {
   global basePortToIfIndex
   set basePortToIfIndex($bp)
}

proc snmpIfBridgeBasePort {if} {
   global ifIndexToBasePort
   set ifIndexToBasePort($if)
}

proc snmpStpInfo {} {
   global snmpS
   set vbl "BRIDGE-MIB!dot1dStpPriority.0
            BRIDGE-MIB!dot1dStpRootCost.0
            BRIDGE-MIB!dot1dStpRootPort.0"
   set vbl [$snmpS get $vbl]
   snmp value $vbl
}

proc snmpStpPortState {pi} {
   global snmpS
   set vb "BRIDGE-MIB!dot1dStpPortState.$pi"
   set vb [$snmpS get $vb]
   snmp value $vb
}

proc snmpLearnedEntryDiscards {} {
   global snmpS
   set vb "BRIDGE-MIB!dot1dTpLearnedEntryDiscards.0"
   set vb [$snmpS get $vb]
   snmp value $vb
}

proc snmpBridgeFdb {} {
   global snmpS
   set bfRoot [mib oid "BRIDGE-MIB!dot1dTpFdbStatus"]
   set vbl "$bfRoot.0
            BRIDGE-MIB!dot1dTpFdbAddress.0
            BRIDGE-MIB!dot1dTpFdbPort.0"
   set fdbList {}
   while {1} {
      set vbl [$snmpS getnext $vbl]
      if {! [mib subtree $bfRoot [snmp oid $vbl 0]]} break
      if {![string compare [snmp value $vbl 0] "learned"]} {
         lappend fdbList [list [snmp value $vbl 2] [snmp value $vbl 1]]
      }
   }
   set fdbList
}
