#!/bin/sh
# $Id: ssScan.tcl,v 1.46 2001/07/24 14:58:54 gdmr Exp $
#
# Copyright 2000 Division of Informatics, University of Edinburgh
# 
# Permission to use, copy, modify, distribute, and sell this software and its
# documentation for any purpose is hereby granted without fee, provided that
# the above copyright notice appear in all copies and that both that
# copyright notice and this permission notice appear in supporting
# documentation, and that the name of the University of Edinburgh not be used
# in advertising or publicity pertaining to distribution of the software
# without specific, written prior permission.  The University of Edinburgh
# makes no representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied warranty.
#
# Comments are invited.  Please send to gdmr@dcs.ed.ac.uk.

# Scan 3Com SuperStack switches, comparing what SNMP claims is
# configured against what's in a "definitive" configuration file.

# expect is in a different place on Solaris and Linux, and we can't
# assume that it's in our PATH as we may be running from a cron job.
# This is the standard tcl hack, only with more than one shell line...
#\
case "`uname -sr`" in
#\
        'SunOS 5'*)     EXPECT=/usr/local/bin/expect ;;
#\
        Linux*)         EXPECT=/usr/bin/expect ;;
#\
         *)             echo "Don't know where to find tclsh" ; exit ;;
#\
esac
#\
exec $EXPECT "$0" -- "$@"
# It's tcl from here on...

# How long should we wait after each switch update?
set afterTime 1500

# Auto-load from the directory that this script lives in
lappend auto_path [file dirname $argv0]

#
# Get the switch configuration name from the command line
# and any flags
#
proc Usage {} {
   global argv0
   puts stderr "$argv0 \[-update] \[-showDown] \[-debug] config-file"
}

set password ""
set updates {}
set vlanUpdates {}
set vlanDeletes {}
set doUpdate 0
set doShowDown 0
set fwdLimit 0
set debug 0
foreach a $argv {
   switch -glob -- $a {
      -update	{ set doUpdate 1 }
      -down     -
      -showDown -
      -showdown { set doShowDown 1 }
      -debug    { set debug 1 }
      -F*       { if {![regexp {^-F([0-9]+)$} $a all fwdLimit]} {
                     Usage ; exit 1 }}
      -P*       { regexp "^-P(.*)$" $a all password }
      --        { if {$debug} {puts "\"--\" ignored"} }
      -*	{ Usage ; exit 1 }
      *		{ if [info exists configFile] { Usage ; exit 1 }
		  set configFile $a }
   }
}

# What's legal in the configuration file?
set confCommands [parseCommands]

# Read the configuration file
puts "Reading configuration file \"$configFile\"..."
if [catch {open $configFile r} cf] {
   puts stderr "Read switch configuration: $cf"
   exit 2
} else {
   if [catch {split [read $cf] "\n"} confFile] {
      puts stderr "Error reading $configFile: $confFile"
      exit 2
   }
   close $cf
}

# Parse it if it read in OK
set bogusCommand 0
foreach confLine $confFile {
   set confLine [string trim $confLine]
   if [string length $confLine] {
      set cfw [lindex [split $confLine] 0]
      if [string compare [string index $cfw 0] "#"] {
         if {[lsearch -exact $confCommands $cfw] >= 0} {
            if [catch {eval $confLine} problem] {
               puts stderr "Error in \"$confLine\""
               if {$debug} {puts stderr "$problem"}
               set bogusCommand 1
            }
         } else {
            puts stderr "Unknown configuration option \"$cfw\""
            set bogusCommand 1
         }
      }
   }
}

# Did it parse?
if {$bogusCommand || ![parseOK]} {
   puts stderr "Terminating due to errors in $configFile"
   exit 2
}

set configured [getPortList]
#puts "Configured: $configured"

snmpInit [getTarget] [getCommunity]
if [catch {snmpDescr} descr] {
   puts "No response from [getTarget]: $descr"
   exit 98
}
if {$descr != "3Com SuperStack II"} {
   puts "[getTarget] doesn't appear to be a 3Com SuperStack II switch"
   puts "Got: [snmpDescr]"
   exit 3
}

proc putDetails {u} {
   set details [snmpStackConfigList $u]
   set d [lindex $details 0]
   set p [lindex $details 2]
   set h [lindex $details 3]
   set s [lindex $details 4]
   set n [lindex $details 5]
   set a [lindex $details 6]
   puts "\t$d, PROM $p, H/W $h, S/W $s"
   puts "\tS/N $n, MAC $a"
}

set since "[clock format [expr [clock seconds] - [snmpUpTime]]]"
puts "Switch [getTarget] ([snmpLocation])"
putDetails 1
puts "\tup since $since"

set t2 [getTarget2]
if {[string length $t2] > 0} {
   snmpInit2 $t2 [getCommunity]
   if [catch {set s2 "[clock format [expr [clock seconds] - [snmpUpTime2]]]"} \
         problem] {
      puts "Target 2 at $t2: $problem"
   } else {
      puts "Target 2 at $t2 up since $s2"
      putDetails 2
   }
}

#
# Bridge forwarding table
#
set led [snmpLearnedEntryDiscards]
if {$led > 0} {
   puts "\t** $led addresses not learned"
}
if {$fwdLimit > 0} {
   puts "Grabbing forwarding database..."
   set bbpList [snmpBridgeBasePorts]
   set fdbList [snmpBridgeFdb]
   foreach f $fdbList {
      set addr [lindex $f 1]
      set bp [lindex $f 0]
      set if [snmpBridgeBasePortIf $bp]
      if [info exists ifAddrs($if)] {
         lappend ifAddrs($if) $addr
      } else {
         set ifAddrs($if) [list $addr]
      }
   }
#  foreach i [lsort -integer [array names ifAddrs]] {
#     puts "$i -> $ifAddrs($i)"
#  }
}

#
# Spanning tree information
#
#puts "Grabbing bridge base ports table..."
snmpBridgeBasePorts
set stpInfo [snmpStpInfo]
set stpPriority [lindex $stpInfo 0]
set stpRootCost [lindex $stpInfo 1]
set stpRootPort [lindex $stpInfo 2]
#puts "STP: prio $stpPriority, rootCost: $stpRootCost, rootPort: $stpRootPort"
if {$stpRootCost > 0} {
   set stpRootBase [snmpBridgeBasePortIf $stpRootPort]
} else {
   set stpRootBase -1
}
#puts "STP: prio $stpPriority, rootCost: $stpRootCost, rootPort: $stpRootPort, rootBase: $stpRootBase"

#
# Verify the VLAN configuration against the switch
#
puts "Scanning for VLANs..."
set vlanIndices [snmpVlanIndices]
puts "VLANs: checking configuration against switch..."
foreach v $vlanIndices {
   set tag [snmpVlanTag $v]
   lappend vlanSwitchTagList $tag
   set confName [getVlanTagOrName $tag]
   if {$confName == ""} {
      puts " - tag $tag ([snmpVlanDescr $v]) on switch but not in configuration"
      lappend vlanDeletes [list ssDeleteVlan [snmpVlanDescr $v] $tag]
   } else {
      set vlanConfAndSwitch($tag) $v
   }
}
if {[llength $vlanDeletes] > 0} {
   puts "** Forcing -showDown due to unconfigured VLANs"
   set doShowDown 1
}
set vlanConfTagList [getVlanTagList]
# Assume that the config file has no all-numeric VLAN names
# (should sanity-check this in the parser)
foreach c $vlanConfTagList {
   if [regexp {^[0-9]+$} $c] {
      if {[lsearch -exact $vlanSwitchTagList $c] < 0} {
         set n [getVlanTagOrName $c]
         puts " - tag $c ($n) in configuration but not on switch"
         lappend updates [list ssCreateVlan $n $c]
      }
   }
}

#
# Get the VLAN stacking information
#
puts "Building VLAN stacking table..."
snmpGrabIfStacking

puts "Scanning for interfaces..."
set actual [snmpEtherIfIndices]
set skipList [getSkipList]

proc displayFwdList {pi} {
   global ifAddrs
   global fwdLimit
   if [info exists ifAddrs($pi)] {
      set n [llength $ifAddrs($pi)]
      if {$n > $fwdLimit} {
         puts "\t-> ($n addresses)"
      } else {
         set text ""
         set wrap 0
         foreach a $ifAddrs($pi) {
            if {[string length $text] > 0} {
               append text ", "
               if {[expr [incr wrap] % 3] == 0} {
                  append text "\n\t   "
               }
            }
            append text "$a"
         }
         puts "\t-> $text"
      }
   }
}

proc displayOne {c portName} {
   global upDown
   global skipList
   global updates
   global vlanUpdates
   global debug
   global fwdLimit
   global stpRootBase
   global stpRootCost
   # Should we ignore it completely?
   if {[lsearch -exact $skipList $c] >= 0} return 
   # Is it up or down?
   if [catch {set ifState [snmpIfUp $c]} problem] {
      puts "[textPort $c]: $portName NOT ON SWITCH STACK?"
      if {$debug} {puts "\t$problem"}
      return
   }
   if {$ifState} {
      set upOrDown "up"
      set upDown($c) 1
   } else {
      set upOrDown "DOWN"
      set upDown($c) 0
   }
   # Get port type, speed, full/half-duplex
   set mauType [snmpMauType $c]
   if {[regexp "dot3MauType(.*)" $mauType all mt]} {
      set mauType $mt
   }
   set mauAuto [snmpMauAuto $c]
   if {$mauAuto == "enabled"} {
      set as "negotiated"
   } else {
      set as "set"
   }
   set pt "[textPort $c]: $portName $upOrDown, $as $mauType"
   # Spanning tree...
   if {$stpRootCost > 0} {
      set stpPS [snmpStpPortState [snmpIfBridgeBasePort $c]]
      append pt ", $stpPS"
      if {$c == $stpRootBase} {
         append pt "/$stpRootCost"
      }
   }
   # Port stacking: look for VLANs and trunks
   set sa [snmpStackAbove $c]
   set trunk 0
   set utv 0
   set tv {}
   if {[llength $sa] > 0} {
      foreach a $sa {
         set type [snmpIfType $a]
         switch -exact -- $type {
            propVirtual {
               # VLAN
               set utv $a
            }
            propMultiplexor {
               # Tagged VLAN, extract the next level
               lappend tv [snmpStackAbove $a]
            }
            ethernetCsmacd {
               # Trunked
               set trunk $a
            }
            default {
               puts stderr "Unexpected ifType $type for $c/$a"
            }
         }
      }
      # Verify untagged VLAN
      #  avn: actual VLAN name
      #  avt: actual VLAN tag
      #  cvn: configuration VLAN name
      #  cvt: configuration VLAN tag
      set avn [snmpVlanDescr $utv]
      set avt [snmpVlanTag $utv]
      set cdv [getPortDefVlan $c]
      set cdt [getVlanTagOrName $cdv]
      append pt ", on $avn ($avt)"
      if {$cdt != "" && $avt != $cdt} {
         set sbm "\tShould be on $cdv ($cdt)"
         lappend updates [list ssDefaultVlan $c $cdt]
      } else {
         set sbm ""
      }
   } else {
      append pt ", on *unknown*"
   }
   puts $pt
   if {[string length $sbm] > 0} {
      puts "$sbm"
   }
   # Compare mauType/mauAuto against the configuration.
   #   should be negotiated + actually negotiated: OK
   #      "   "    "   "    + actually set: complain, queue correction
   #   should be set + actually negotiated: complain, queue correction
   #      "   "   "  + actually set: check speed, complain and queue
   #                                 correction if necessary
   set confSpeed [getPortSpeed $c]
   if {$confSpeed == "-"} {
      if {$mauAuto == "enabled"} {
         if {$debug} {puts "\tSpeed set and configured auto"}
      } else {
         puts "\tSpeed should be auto-negotiated"
         lappend updates [list ssSetAuto $c]
      }
   } else {
      if {$mauAuto =="enabled"} {
         puts "\tSpeed should be set to $confSpeed"
         lappend updates [list ssSetSpeed $c $confSpeed]
      } else {
         splitSpeed $confSpeed cs ca
         splitSpeed $mauType ms ma
         if {$cs == $ms && $ca == $ma} {
            if {$debug} {puts "\tSpeed set $mauType and configured $confSpeed"}
         } else {
            puts "\tSpeed should be set to $confSpeed"
            lappend updates [list ssSetSpeed $c $confSpeed]
         }
      }
   }
   # Is it trunked?
   if {$trunk > 0} {
      puts "\tAlso on [snmpIfDescr $trunk]"
   }
   # What should be there?
   set vc [getPortVlans $c]
   # Got the names, get the tags
   set tc {}
   foreach v $vc {
      lappend tc [getVlanTagOrName $v]
   }
   # List of tags seen on the switch
   set seen {}
   # Tagged VLANs: compare the configured list
   # against the actual one from the switch
   if {[llength $tv] > 0} {
      # Text for: tagged present; extra present
      set tText ""
      set eText ""
      # Wrap counters for above.  Allow for descriptive
      # text length
      set tWrap 0
      set eWrap 1
      # First scan the list of things that are present.
      # For each: get tag and description
      #           check configured list and note accordingly
      foreach t $tv {
         set vt [snmpVlanTag $t]
         set vd [snmpVlanDescr $t]
         lappend seen $vt
         if {[lsearch -exact $tc $vt] >= 0} {
            # Configured
            if {[string length $tText] > 0} {
               append tText ", "
               if {[expr [incr tWrap] & 3] == 0} {
                  append tText "\n\t\t"
               }
            }
            append tText "$vd ($vt)"
         } else {
            # Not a configured one
            if {[string length $eText] > 0} {
               append eText ", "
               if {[expr [incr eWrap] & 3] == 0} {
                  append eText "\n\t\t"
               }
            }
            append eText "$vd ($vt)"
            lappend vlanUpdates [list ssRemoveVlan $c $vt]
         }
      }
      if {[string length $tText] > 0} {
         puts "\tTagged VLANs: $tText"
      }
      if {[string length $eText] > 0} {
         puts "\tExtra tagged VLANs: $eText"
      }
   }
   # Now check the list of what should be there against
   # what we've found on the switch
   set mText ""
   set mWrap 1
   foreach t $tc {
      if {[lsearch -exact $seen $t] < 0} {
         # Not there
         if {[string length $mText] > 0} {
            append mText ", "
            if {[expr [incr mWrap] & 3] == 0} {
               append mText "\n\t\t"
            }
         }
         append mText "[getVlanTagOrName $t] ($t)"
         lappend updates [list ssAddTaggedVlan $c $t]
      }
   }
   if {[string length $mText] > 0} {
      puts "\t** Missing tagged VLANs: $mText"
   }
   # Port statistics
   set stats [snmpIfStats $c]
   set st "\tFrames in: [lindex $stats 1]"
   set e [lindex $stats 2]
   if {$e != 0} {
      append st " ($e error"
      if {$e != 1} {append st "s"}
      append st ")"
   }
   append st ", out: [lindex $stats 4]"
   set e [lindex $stats 5]
   if {$e != 0} {
      append st " ($e error"
      if {$e != 1} {append st "s"}
      append st ")"
   }
   set ase [snmpMauAvailableExits $c]
   if {$ase > 0} {
      append st "; link down: $ase"
   }
   puts "$st"
   # Port forwarding table
   if {$fwdLimit > 0} {
      displayFwdList $c
   }
}

# Process everything from the configuration file
# (i.e. that should be there)
foreach c $configured {
   displayOne $c [getPortName $c]
}

# Now what's actually there, skipping things that
# we've already noted from the configuration file
foreach a $actual {
   if {! [info exists upDown($a)]} {
      if {$doShowDown || [snmpIfUp $a]} {
         displayOne $a "*unknown peer*"
      }
   }
}

# Finally, should we update the switch to match the configuration?
# Is there anything to do?
set updates [concat $vlanUpdates $updates $vlanDeletes]
if {($doUpdate || $debug) && ([llength $updates] > 0)} {
   if {$debug} {
      puts "Switch updates required:"
      foreach uc $updates {
         puts "   $uc"
      }
   }
   if {$doUpdate} {
      set eCount 0
      if [ssConnect [getTarget] $password] {
         foreach uc $updates {
            if {[eval $uc]} {
               after $afterTime
            } else {
               incr eCount
            }
         }
         ssDisconnect
         if {$eCount} {
            puts "$eCount error(s)"
            exit 7
         }
         exit 0
      } else {
         puts "Couldn't connect"
         exit 7
      }
   }
   exit 0
}
